/*
 * A page mapper providing memory pages to satisfy file accesses, masking the
 * limits of a visible region of the file's contents.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <string.h>

#include "masked_page_mapper.h"



/* Provide mapped pages populated using the given 'mapper', with a visible
   region defined that causes the limits of this region to be masked. */

MaskedPageMapper::MaskedPageMapper(GenericPageMapper *mapper,
                                   offset_t start_visible, offset_t end_visible)
: _mapper(mapper), _start_visible(start_visible), _end_visible(end_visible)
{
    _size = _mapper->get_data_size();
}

MaskedPageMapper::~MaskedPageMapper()
{
    if (_start_flexpage.region != NULL)
        _memory.release(_start_flexpage.region);

    if (_end_flexpage.region != NULL)
        _memory.release(_end_flexpage.region);

    if (_zero_flexpage.region != NULL)
        _memory.release(_zero_flexpage.region);
}

/* Interface for the pager. */

/* Return a flexpage providing access to the indicated file 'offset'.

   The returned flexpage will either provide direct access to the underlying
   file content or be a completely new flexpage containing masked content. */

Flexpage *MaskedPageMapper::get(offset_t offset, map_flags_t flags)
{
    Flexpage *f = _mapper->get(offset, flags);
    Flexpage *mf = get_masked_flexpage(f);

    if (f != mf)
        _mapper->queue(f);

    return mf;
}

/* Queue the given 'flexpage'. */

void MaskedPageMapper::queue(Flexpage *flexpage)
{
    if (dynamic_cast<MaskedFlexpage *>(flexpage) == NULL)
        _mapper->queue(flexpage);
}

/* Flush pages in the given range from 'start' with 'size'. */

void MaskedPageMapper::flush_all(offset_t start, offset_t size)
{
    /* NOTE: This might be superfluous since masked regions probably should be
             read-only or copy-on-write. */

    _mapper->flush_all(start, size);
}

/* Return the maximum extent of the mapped resource. */

offset_t MaskedPageMapper::get_data_size()
{
    return _size;
}

/* Set the maximum extent of the mapped resource. */

void MaskedPageMapper::set_data_size(offset_t size)
{
    _size = size;
}

/* Internal flexpage retrieval methods. */

/* Detect flexpages with masked content, introducing separate flexpages
   offering masked content from the same region. */

Flexpage *MaskedPageMapper::get_masked_flexpage(Flexpage *flexpage)
{
    /* Determine whether the flexpage involves the limits of the visible
       region or is beyond such limits. */

    bool has_start = flexpage->supports_position(_start_visible) &&
                     flexpage->base_offset != _start_visible;
    bool has_end = flexpage->supports_position(_end_visible);
    bool has_zero = (flexpage->base_offset >= _end_visible) ||
                    (flexpage->base_offset + flexpage->size < _start_visible);

    /* Return the original flexpage within the visible limits. */

    if (!has_start && !has_end && !has_zero)
        return flexpage;

    /* Allocate and populate a region in one of the preallocated flexpages for
       masked content. */

    Flexpage &masked = has_start ? _start_flexpage :
                       has_end ? _end_flexpage :
                                 _zero_flexpage;

    allocate_region(flexpage, masked);
    populate_region(flexpage, masked, has_start, has_end);

    /* Associate the preallocated flexpage with the original flexpage. */

    flexpage->associate(&masked);

    return &masked;
}

void MaskedPageMapper::allocate_region(Flexpage *flexpage, Flexpage &masked)
{
    offset_t needed = flexpage->region->size();

    /* Attempt to re-use an existing region. */

    if (masked.region != NULL)
    {
        if (masked.region->size() == needed)
            return;
        else
            _memory.release(masked.region);
    }

    /* Set the region in the flexpage. */

    masked.set_region(_memory.region(needed));
    masked.reset(flexpage->base_offset);
}

void MaskedPageMapper::populate_region(Flexpage *flexpage, Flexpage &masked,
                            bool has_start, bool has_end)
{
    /* Without any limits involved, provide a zero flexpage. */

    if (!has_start && !has_end)
    {
        memset((void *) masked.region->start, 0, masked.size);
        return;
    }

    /* Determine the content limits given start and/or end offsets. */

    offset_t start_offset = has_start ? _start_visible - masked.base_offset : 0;
    offset_t end_offset = has_end ? _end_visible - masked.base_offset : masked.size;

    if (has_start)
        memset((void *) masked.region->start, 0, start_offset);

    memcpy((void *) (masked.region->start + start_offset),
           (const void *) (flexpage->region->start + start_offset),
           end_offset - start_offset);

    if (has_end)
        memset((void *) (masked.region->start + end_offset), 0, masked.size - end_offset);
}



/* Masked flexpage constructor. */

MaskedFlexpage::MaskedFlexpage(Region *region)
: Flexpage(region)
{
}

// vim: tabstop=4 expandtab shiftwidth=4
